from functools import cached_property

from dataclasses import dataclass
from itertools import chain
from typing import Iterable

from parse_qwantz.box import Box
from parse_qwantz.colors import Color
from parse_qwantz.fonts import Font, CharBox
from parse_qwantz.lines import get_line
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage


GREEK_ALPHABET = "αάβγδεέζηήθιίκλμνξοόπρσςτυύφχψωώ"
LATIN_TO_GREEK = {
    "v": "ν",
    "o": "ο",
    "ó": "ό",
}


@dataclass(frozen=True, eq=True)
class TextLine:
    char_boxes: list[CharBox]
    font: Font
    color: Color

    def __repr__(self):
        return f"TextLine({repr(self.content)}, {self.font.name})"

    def __str__(self):
        return self.content

    @cached_property
    def start(self) -> Pixel:
        return self.char_boxes[0].box.top_left

    @cached_property
    def end(self) -> Pixel:
        return Pixel(
            x=self.char_boxes[-1].box.right + len(self.font.italic_offsets),
            y=max(char_box.box.bottom for char_box in self.char_boxes)
        )

    @cached_property
    def content(self) -> str:
        content = ''.join(char_box.char for char_box in self.char_boxes)
        return content

    @cached_property
    def is_bold(self) -> bool:
        return all(char_box.is_bold for char_box in self.char_boxes)

    @cached_property
    def is_italic(self) -> bool:
        return bool(self.font.italic_offsets)

    @property
    def contains_bold(self) -> bool:
        return any(char_box.is_bold for char_box in self.char_boxes if char_box.char != ' ')

    def box(self, padding: int = 0) -> Box:
        return Box(self.start, self.end).with_margin(padding, padding)

    def base_box(self, padding: int = 0) -> Box:
        _, y0 = self.start
        x1, _ = self.end
        y1 = y0 + self.font.base
        return Box(self.start, Pixel(x1, y1)).with_margin(padding, padding)

    def find_pixel(self) -> Pixel:
        return min(self.char_boxes[0].pixels)

    def __hash__(self):
        return id(self)


def try_text_line(start: Pixel, image: SimpleImage, font: Font) -> tuple[TextLine, list[str]] | None:
    x0, y0 = start
    if font.italic_offsets:
        max_x_offset = font.space_width - 3
    else:
        max_x_offset = 0
    max_y_offset = font.height - 1
    for x in range(x0 - max_x_offset, x0 + 1):
        for y in range(y0, y0 - max_y_offset - 1, -1):
            if result := get_text_line(Pixel(x, y), image, font):
                return result


def get_text_line(start: Pixel, image: SimpleImage, font: Font) -> tuple[TextLine, list[str]] | None:
    warnings = []
    char_box, complement = font.get_char(start, image=image, is_first=True)
    if char_box is None or char_box.char == ' ':
        return None
    char_boxes = [char_box]
    spaces = []
    is_bold = char_box.is_bold
    is_italic = char_box.is_italic
    x, y = start
    while True:
        x = char_box.box.right
        if x >= image.width:
            break
        char_box, complement = font.get_char(Pixel(x, y), image, first_column=complement)
        inline_offset_warning = None
        if char_box is None and not (len(char_boxes) == 1 and char_boxes[0].char == "'"):
            for offset in ((0, -1), (0, 1)):
                off_x, off_y = offset
                char_box, complement = font.get_char(Pixel(x + off_x, y + off_y), image)
                if char_box and char_box.char == ' ':
                    char_box = None
                if char_box is not None and char_box.char not in "_'":
                    content_so_far = ''.join(char_box.char for char_box in char_boxes)
                    inline_offset_warning = (
                        f"Inline offset after {content_so_far + ' '*len(spaces)!r},"
                        f" before {char_box.char!r}: {(off_x, off_y)}"
                    )
                    x += off_x
                    y += off_y
                    break
        if char_box is None:
            break
        # this is to avoid treating "i"/"j" dots as periods, but still allow lines starting with "..."
        if len(char_boxes) == 1 and char_boxes[0].char == '.' and char_box.char != '.':
            return None
        if char_box.char in "'|-":
            any_pixel = next(iter(char_box.pixels))
            result = get_line(any_pixel, image)
            if result:
                line, _pixels, _width = result
                if (
                    min(line[0].x, line[1].x) < char_box.box.left
                    or min(line[0].y, line[1].y) < char_box.box.top
                    or max(line[0].x, line[1].x) > char_box.box.right
                    or max(line[0].y, line[1].y) > char_box.box.bottom
                ):
                    break
        if inline_offset_warning:
            warnings.append(inline_offset_warning)
        if char_boxes:
            previous_char = char_boxes[-1].char
            if previous_char in GREEK_ALPHABET and char_box.char in LATIN_TO_GREEK:
                char_box = char_box.with_char(LATIN_TO_GREEK[char_box.char])
            elif previous_char in LATIN_TO_GREEK and char_box.char in GREEK_ALPHABET:
                char_boxes[-1] = char_boxes[-1].with_char(LATIN_TO_GREEK[previous_char])
        if char_box.char == ' ':
            if len(char_boxes) == 1 and char_boxes[0].char == "'":
                return
            spaces.append(
                CharBox.space(is_bold, is_italic, char_box.box)
            )
            if len(spaces) > 2:
                break
        else:
            if spaces:
                char_boxes.extend(spaces)
                spaces = []
            char_boxes.append(char_box)
            is_bold = char_box.is_bold
            is_italic = char_box.is_italic
    first_char = char_boxes[0].char
    if (
        len(char_boxes) == 1 and not first_char.isalnum()
        and first_char not in '!?$'
        and not (first_char == '-' and font.group == 'LC13')
        and not (first_char == '/' and font.name == "Regular" and not is_bold and not is_italic)
    ):
        return
    if len(char_boxes) >= 2 and all(char_box.char in " \",.'‘’“”|-/·•" for char_box in char_boxes):
        if ''.join(char_box.char for char_box in char_boxes) != "...":
            return
    char_boxes = list(adjust_spaces(char_boxes))
    if len(char_boxes) > 2 and char_boxes[-1].char == "'" and char_boxes[-2].char == " ":
        char_boxes = char_boxes[:-2]
    color = image.get_pixel(min(char_boxes[0].pixels))
    return TextLine(char_boxes, font, color), warnings


def adjust_spaces(char_boxes: list[CharBox]) -> Iterable[CharBox]:
    for char_box, next_char_box in zip(char_boxes, chain(char_boxes[1:], [None])):
        if char_box.char == ' ' and char_box.box.right > next_char_box.box.left:
            new_box = Box(char_box.box.top_left, Pixel(next_char_box.box.left, char_box.box.bottom))
            if new_box.width > 2:
                yield char_box.with_box(new_box)
        else:
            yield char_box


def cleanup_text_lines(text_lines: list[TextLine]) -> list[TextLine]:
    return sorted(
        text_lines,
        key=lambda l: (l.start.y, l.start.x)
    )


def group_text_lines(text_lines: list[TextLine]) -> list[list[TextLine]]:
    grouped_text_lines = []
    used: set[TextLine] = set()
    for text_line in text_lines:
        if text_line in used:
            continue
        used.add(text_line)
        group = [text_line]
        for other_text_line in text_lines:
            if other_text_line in used:
                continue
            if other_text_line.font.group != text_line.font.group:
                continue
            box = group[-1].base_box()
            other_box = other_text_line.base_box()
            vertical_offset = abs(box.bottom - other_box.bottom)
            same_font = other_text_line.font == text_line.font
            if (vertical_offset <= 1 and same_font) or vertical_offset == 0:
                distance = other_box.left - box.right
                width = max(group[-1].font.space_width, other_text_line.font.space_width)
                max_distance = width * 3
                if -1 <= distance <= max_distance:
                    group.append(other_text_line)
                    used.add(other_text_line)
        grouped_text_lines.append(group)
    return grouped_text_lines



from functools import reduce
from logging import getLogger

from parse_qwantz.box import Box
from parse_qwantz.fonts import ALL_FONTS
from parse_qwantz.lines import Line, get_line
from parse_qwantz.match_lines import Character
from parse_qwantz.text_lines import TextLine, try_text_line, cleanup_text_lines
from parse_qwantz.detect_thought import get_thought
from parse_qwantz.pixels import Pixel, remove_subsequence
from parse_qwantz.shape import get_shape
from parse_qwantz.simple_image import SimpleImage

logger = getLogger()


def get_elements(
    image: SimpleImage
) -> tuple[list[Line], list[int], list[Box], list[TextLine], list[Character], list[list[Pixel]]]:
    text_lines: list[TextLine] = []
    lines: list[Line] = []
    line_widths: list[int] = []
    thoughts: list[Box] = []
    unmatched: list[list[Pixel]] = []
    sorted_pixels = sorted(image.pixels)
    extra_characters = []
    while sorted_pixels:
        pixel = sorted_pixels[0]
        pixels = set(sorted_pixels)
        tmp_image = SimpleImage(
            width=image.width,
            height=image.height,
            pixels={pixel: color for pixel, color in image.pixels.items() if pixel in pixels},
        )
        text_line_candidates = (try_text_line(pixel, tmp_image, font) for font in ALL_FONTS)
        text_line_candidates = [text_line for text_line in text_line_candidates if text_line]
        longest_candidate = max(text_line_candidates, key=lambda tl: tl[0].box().right, default=None)
        # UGLY SPECIAL CASE AHOY
        if longest_candidate and longest_candidate[0].content == "-" and longest_candidate[0].font.name == "Italic":
            try:
                longest_candidate = next(c for c in text_line_candidates if c[0].font.name == "Regular")
            except StopIteration:
                pass
        if longest_candidate:
            longest_line, warnings = longest_candidate
            for warning in warnings:
                logger.warning(warning)
            text_lines.append(longest_line)
            pixels = reduce(set.union, (char_box.pixels for char_box in longest_line.char_boxes))
            sorted_pixels = remove_subsequence(sorted_pixels, sorted(pixels))
        else:
            result = get_line(pixel, tmp_image)
            if result:
                line, line_pixels, width = result
                lines.append(line)
                line_widths.append(width)
                sorted_pixels = remove_subsequence(sorted_pixels, line_pixels)
            elif result := get_batman(pixel, tmp_image):
                batman_box, batman_pixels = result
                extra_characters.append(Character("Floating Batman head", (batman_box,)))
                sorted_pixels = remove_subsequence(sorted_pixels, batman_pixels)
            elif result := get_thought(pixel, tmp_image):
                box, thought_pixels = result
                thoughts.append(box)
                sorted_pixels = remove_subsequence(sorted_pixels, thought_pixels)
            else:
                unmatched_pixels = sorted(get_shape(pixel, tmp_image))
                unmatched.append(unmatched_pixels)
                sorted_pixels = remove_subsequence(sorted_pixels, unmatched_pixels)
                logger.warning(f"No match found for shape at {(pixel.x, pixel.y)} ({len(unmatched_pixels)} pixels)")
    return lines, line_widths, thoughts, cleanup_text_lines(text_lines), extra_characters, unmatched


def get_batman(pixel: Pixel, image: SimpleImage) -> tuple[Box, list[Pixel]] | None:
    pixels = get_shape(pixel, image)
    if len(pixels) != 187:
        return None
    if Pixel(pixel.x + 11, pixel.y - 7) in pixels and Pixel(pixel.x + 11, pixel.y - 8) not in pixels:
        return Box(Pixel(pixel.x, pixel.y - 7), Pixel(pixel.x + 14, pixel.y + 11)), sorted(pixels)
    if Pixel(pixel.x + 2, pixel.y - 17) in pixels and Pixel(pixel.x + 2, pixel.y - 18) not in pixels:
        return Box(Pixel(pixel.x, pixel.y - 7), Pixel(pixel.x + 14, pixel.y + 11)), sorted(pixels)
    return None



import logging
from collections.abc import Iterable

from importlib.resources import as_file, files
from itertools import chain

import parse_qwantz

logger = logging.getLogger()


def make_word_set(dict_path: str, extra_words: Iterable[str] = ()) -> frozenset[str]:
    with as_file(files(parse_qwantz).joinpath(dict_path)) as dict_path:
        dict_words = (line.rstrip('\n').lower() for line in open(dict_path))
    return frozenset(chain(dict_words, extra_words))


QWANTZ_WORD_SET = (
    make_word_set('dict/unambiguous-qwantz.txt')
    | make_word_set('dict/html-words.txt')
    | make_word_set('dict/manual-additions.txt')
) - make_word_set('dict/manual-removed.txt')


def disambiguate_hyphen(part1: list[str], part2: list[str]):
    word1 = part1[-1]
    word2 = part2[0]
    if not (word1 and word2):
        return False
    if word1[-1].islower() and word2[0].isupper():
        return True
    if word2[0].isdigit():
        return True
    word1_lower = word1.lower()
    word2_lower = word2.lower()

    no_hyphen = word1_lower + word2_lower in QWANTZ_WORD_SET
    with_hyphen = f"{word1_lower}-{word2_lower}" in QWANTZ_WORD_SET
    all_with_hyphen = f"{'-'.join(part1)}-{'-'.join(part2)}" in QWANTZ_WORD_SET

    parts_for_logging = f"{'-'.join(part1)}/{'-'.join(part2)}"
    if no_hyphen:
        if with_hyphen or all_with_hyphen:
            logger.warning(f"Ambiguous hyphen ({parts_for_logging}); both in Qwantz dict")
        return False
    if with_hyphen or all_with_hyphen:
        return True
    if len(part1) > 1 or len(part2) > 1:
        if word1_lower not in QWANTZ_WORD_SET or word2_lower not in QWANTZ_WORD_SET:
            logger.warning(f"Unexpected hyphenation in multi-hyphened phrase ({parts_for_logging})")
        return True
    logger.warning(f"Unresolved hyphen ({parts_for_logging})")
    return False



import logging
from typing import Iterable

from parse_qwantz.text_blocks import TextBlock
from parse_qwantz.match_lines import Target, Character, OFF_PANEL
from parse_qwantz.text_lines import TextLine

logger = logging.getLogger()

MULTI_OFF_PANEL = Character.from_name("multiple off-panel voices")


def match_blocks(
    line_matches: Iterable[tuple[Target, Target]], text_blocks: list[TextBlock]
) -> tuple[dict[TextBlock, list[Character]], list[TextBlock], list[tuple[TextLine, TextLine]]]:
    block_matches: dict[TextBlock, tuple[list[Character], TextLine]] = {}
    blocks_by_line: dict[TextLine, TextBlock] = {line: block for block in text_blocks for line in block.lines}
    neighbors: list[tuple[TextLine, TextLine]] = []
    for target1, target2 in line_matches:
        if isinstance(target1, TextLine) and isinstance(target2, TextLine):
            neighbors.append((target1, target2))
        else:
            character, line = (target1, target2) if isinstance(target1, Character) else (target2, target1)
            line: TextLine
            block = blocks_by_line[line]
            if block not in block_matches:
                block_matches[block] = ([character], line)
            else:
                prev_characters, prev_line = block_matches[block]
                if not block.can_split(prev_line, line):
                    if len(prev_characters) == 2:
                        logger.warning("More than two characters share a line")
                    if prev_line != line:
                        logger.warning("Shared separate lines")
                    if character == OFF_PANEL and OFF_PANEL in prev_characters:
                        prev_characters.remove(OFF_PANEL)
                        new_characters = prev_characters + [MULTI_OFF_PANEL]
                    elif character == OFF_PANEL and MULTI_OFF_PANEL in prev_characters:
                        new_characters = prev_characters
                    else:
                        new_characters = prev_characters + [character]
                    block_matches[block] = (new_characters, line)
                else:
                    block1, block2, alignment = block.split(prev_line, line)
                    if alignment.no_gap and (alignment.left_aligned or alignment.char_aligned):
                        logger.warning(
                            f"Splitting blocks: [{block1.content()[:12]}], [{block2.content()[:12]}],"
                            f" strength: {alignment.strength}"
                        )
                    for b1line in block1.lines:
                        blocks_by_line[b1line] = block1
                    for b2line in block2.lines:
                        blocks_by_line[b2line] = block2
                    del block_matches[block]
                    block_matches[block1] = (prev_characters, prev_line)
                    block_matches[block2] = ([character], line)
    while neighbors:
        neighbors_left = []
        for line1, line2 in neighbors:
            if blocks_by_line[line1] == blocks_by_line[line2]:
                # no longer happens
                block = blocks_by_line[line1]
                if block.row_index(line1) == block.row_index(line2):
                    logger.warning(f"Line connects two text lines in one row: {line1.content} -- {line2.content}")
                    continue
                block1, block2, _ = block.split(line1, line2)
                logger.warning(
                    f"Splitting blocks from one character: [{block1.content()}], [{block2.content()}]"
                )
                for b1line in block1.lines:
                    blocks_by_line[b1line] = block1
                for b2line in block2.lines:
                    blocks_by_line[b2line] = block2
                if block in block_matches:
                    characters, line = block_matches[block]
                    del block_matches[block]
                    block_matches[block1] = (characters, line)
                    block_matches[block2] = (characters, line)
            else:
                block1 = blocks_by_line[line1]
                block2 = blocks_by_line[line2]
                if (
                    block1.row_index(line1) not in (0, len(block1.rows) - 1)
                    or block2.row_index(line2) not in (0, len(block2.rows) - 1)
                ):
                    if not (line1.content.isupper() and line2.content.isupper() and line1.is_bold and line2.is_bold):
                        logger.warning(f"Matching a non-edge line to another line: {line1.content} -- {line2.content}")
            if block1 in block_matches and block2 in block_matches:
                if block1.start.y < block2.start.y:
                    first_block, second_block = block1, block2
                else:
                    first_block, second_block = block2, block1
                characters, line = block_matches[first_block]
                other_characters, other_line = block_matches[second_block]
                block_matches[second_block] = characters + other_characters, other_line
            elif block1 in block_matches:
                block_matches[block2] = block_matches[block1][0], line2
            elif block2 in block_matches:
                block_matches[block1] = block_matches[block2][0], line1
            else:
                neighbors_left.append((line1, line2))
        if neighbors == neighbors_left:
            logger.warning(f"Unmatched connected text lines: {neighbors}")
            break
        neighbors = neighbors_left
    final_block_matches = {block_id: characters for block_id, (characters, line) in block_matches.items()}
    text_blocks = list(set(blocks_by_line.values()))
    return final_block_matches, text_blocks, neighbors



import logging
from functools import cache
from importlib.resources import as_file, files

from PIL import Image

import parse_qwantz
from parse_qwantz.colors import square_distance, COLOR_THRESHOLD
from parse_qwantz.pixels import normalize_color

logger = logging.getLogger()

DIM = (735, 500)

MASK_FILE_PATH = files(parse_qwantz).joinpath('img/mask.png')

SAMPLE = [
    ((113, 183), (128, 255, 64)),
    ((290, 160), (255, 128, 161)),
    ((704, 183), (255, 242, 179)),
    ((143, 375), (255, 191, 82)),
    ((452, 405), (255, 191, 82)),
    ((585, 386), (69, 196, 5)),
]


class ImageError(Exception):
    pass


@cache
def get_mask_image() -> Image:
    with as_file(MASK_FILE_PATH) as image_path:
        return Image.open(image_path)


def prepare_image(image: Image) -> tuple[Image, list[int]]:
    if image.size != DIM:
        logger.error(f"Wrong image dimensions: {image.size}, only {DIM} is valid")
        raise ImageError(f"Wrong image dimensions: {image.size}, only {DIM} is valid")
    palette = image.getpalette()
    palette = tuple(palette) if palette else None
    good_panels: list[int] = []
    for panel_no, (pixel, expected_color) in enumerate(SAMPLE, start=1):
        color = image.getpixel(pixel)
        color = normalize_color(color, palette)
        if square_distance(color, expected_color) <= COLOR_THRESHOLD:
            good_panels.append(panel_no)
        else:
            logger.info(f"Invalid template: expected {expected_color} at {pixel}; found {color}")
    if not good_panels:
        logger.error("Invalid template")
        raise ImageError(f"Invalid template")
    all_white = Image.new(mode='RGB', size=DIM, color=(255, 255, 255))
    return Image.composite(image, all_white, get_mask_image()), good_panels



import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib.resources import as_file, files
from itertools import islice, chain
from pathlib import Path
from typing import NamedTuple, ContextManager, Iterator, ForwardRef, Union, Any, Sequence

from PIL import Image

import parse_qwantz
from parse_qwantz.box import Box
from parse_qwantz.char_variants import VARIANTS
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage

CHARS = (
    "0123456789"
    "abcdefghijklmnopqrstuvwxyz"
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    """!"#$%&'()*+,-./:;<=>?@[]^{|}‘’“”·•™"""
    "éÉïçêáÁèàßûñùÜóÆÖИÇâæÊöü"
    "°®♪♫±_~¡²«»∩×"
    "αάβγδεέζηήθιίκλμνξοόπρσςτυύφχψωώˣ"
)
FORBIDDEN_FIRST_CHARS = "%&)=^|}”·™_"

ACCEPT = -1

FONT_SIZES = [
    (13, 'Regular', 8, 2, 1, None),
    (12, 'Condensed', 7, 0, 0, None),
    (11, 'Small', 7, 0, 0, "small"),
    (10, 'Petite', 6, 0, 0, "small"),
    (9, 'Mini', 5, 0, 0, "tiny"),
    (8, 'Tiny', 5, 0, 0, "tiny"),
]


@dataclass
class CharInfo:
    char: str
    left_padding: int
    right_padding: int
    extra_info: str | None = None


class CharBox(NamedTuple):
    char: str
    box: Box
    is_bold: bool
    is_italic: bool
    pixels: set[Pixel]
    extra_info: str | None = None

    def with_box(self, box: Box):
        return CharBox(self.char, box, self.is_bold, self.is_italic, self.pixels)

    def with_char(self, char: str):
        return CharBox(char, self.box, self.is_bold, self.is_italic, self.pixels)

    @classmethod
    def space(cls, is_bold: bool, is_italic: bool, box: Box | None = None) -> "CharBox":
        return cls(
            char=' ',
            box=box or Box.dummy(),
            is_bold=is_bold,
            is_italic=is_italic,
            pixels=set(),
        )


FSA_BACKREF = ForwardRef("FSA")

FSA = dict[int, Union[FSA_BACKREF, CharInfo]]


@dataclass
class Font(ABC):
    name: str
    space_width: int
    height: int
    base: int
    automaton: FSA
    initial_padding: int
    final_padding: int
    is_bold: bool
    italic_offsets: set[int]
    group: str
    max_cut_bottom: int
    max_cut_top: int
    display_name: str | None = None
    is_mono: bool = True

    base_left_padding: int = 0
    base_right_padding: int = 0
    skip_chars: Sequence[str] = ()

    def get_char(
        self,
        pixel: Pixel,
        image: SimpleImage,
        is_first: bool = False,
        first_column: int | None = None,
    ) -> tuple[CharBox | None, int | None]:
        char_box, complement = self._get_char_box_from_columns(
            pixel, self._get_columns(pixel, image), is_first, first_column
        )
        if char_box:
            return char_box, complement

        for cut_bottom in range(1, self.max_cut_bottom + 1):
            columns = self._get_columns(pixel, image, cut_bottom=cut_bottom)
            char_box, complement = self._get_char_box_from_columns(pixel, columns, is_first, first_column)
            if char_box:
                return char_box, complement

        for cut_top in range(1, self.max_cut_top + 1):
            columns = self._get_columns(pixel, image, cut_top=cut_top)
            char_box, complement = self._get_char_box_from_columns(pixel, columns, is_first, first_column)
            if char_box:
                return char_box, complement
        return None, None

    def _get_char_box_from_columns(
        self, pixel: Pixel, columns: Iterator[tuple[int, int]], is_first: bool, first_column: int | None
    ) -> tuple[CharBox | None, int | None]:
        is_italic = bool(self.italic_offsets)
        if first_column is not None:
            pixel = Pixel(pixel.x - 1, pixel.y)
            columns = chain([(pixel.x, first_column)], columns)
        x = column = 0  # to satisfy linters; columns are never empty
        for x, column in islice(columns, self.initial_padding + 1):
            if column != 0:
                break
        else:
            for x, column in columns:
                if column != 0:
                    char_box = CharBox.space(
                        self.is_bold,
                        is_italic,
                        Box(pixel, Pixel(x, pixel.y + self.height)),
                    )
                    return char_box, None
        x0 = x
        state = self.automaton
        accepted = None
        char_columns = []
        for x, column in chain([(x0, column)], columns):
            if column not in state:
                if not self.is_mono and len(state) == 1:
                    actual_column, next_state = next(iter(state.items()))
                    if actual_column != ACCEPT and ACCEPT in next_state and column | actual_column == column:
                        complement = column & ~actual_column
                        if complement in self.automaton:
                            char_columns.append(actual_column)
                            accepted = (x, next_state[ACCEPT], complement, char_columns)
                if accepted:
                    break
                return None, None
            state = state[column]
            char_columns.append(column)
            if ACCEPT in state:
                accepted = (x, state[ACCEPT], None, list(char_columns))
        else:
            if not accepted:
                return None, None
        x, char_info, complement, char_columns = accepted
        if is_first and char_info.char in FORBIDDEN_FIRST_CHARS:
            return None, None
        if char_info.left_padding < 0 and x0 - char_info.left_padding - pixel.x >= self.initial_padding:
            return CharBox.space(
                self.is_bold,
                is_italic,
                Box(pixel, Pixel(x0 - char_info.left_padding, pixel.y + self.height)),
            ), char_columns[0]
        pixels = get_pixels_from_columns(char_columns, self.height, x0, pixel.y, self.italic_offsets)
        if char_info.right_padding < 0 and complement is None:
            complement = 0
        return CharBox(
            char_info.char,
            Box(
                Pixel(x0 - char_info.left_padding, pixel.y),
                Pixel(x + 1, pixel.y + self.height),
            ),
            self.is_bold,
            is_italic,
            pixels,
            char_info.extra_info,
        ), complement

    def _get_columns(
        self, pixel: Pixel,
        image: SimpleImage,
        cut_bottom: int = 0,
        cut_top: int = 0,
    ) -> Iterator[tuple[int, int]]:
        max_x = image.width + self.final_padding
        if self.is_mono:
            max_x = min(max_x, pixel.x + self.space_width * 2 + 1)
        for x in range(pixel.x, max_x):
            yield x, self._get_column(x, pixel.y, image, cut_bottom, cut_top)

    def _get_column(self, x: int, y: int, image: SimpleImage, cut_bottom: int, cut_top: int) -> int:
        return get_column(x, y, image, self.height, self.italic_offsets, cut_bottom, cut_top)

    def __str__(self):
        return self.name

    def __repr__(self):
        return (
            f"Font(name={self.name}, height={self.height}, "
            f"is_bold={self.is_bold}, is_italic={bool(self.italic_offsets)})"
        )

    @classmethod
    def from_file(
        cls,
        file_path_context_manager: ContextManager[Path],
        name: str,
        width: int,
        italic_offsets: set[int],
        is_bold: bool,
        group: str,
        max_cut_bottom: int,
        max_cut_top: int,
        display_name: str | None = None,
        **kwargs: Any,
    ) -> "Font":
        with file_path_context_manager as file_path:
            image = SimpleImage.from_image(Image.open(file_path))
        height = image.height
        automaton: FSA = {}
        accepting_states: list[CharInfo] = []
        base = 0  # unused value
        for i, (char, columns) in enumerate(zip(CHARS, cls.get_input_columns(image, width, italic_offsets, is_bold))):
            if char in cls.skip_chars:
                continue
            if char == '0':
                for y in range(height - 1, 0, -1):
                    if any((1 << (height - 1 - y)) & column for column in columns):
                        base = y
                        break
            if maybe_char_info := cls.update_automaton(char, columns, automaton):
                accepting_states.append(maybe_char_info)
        for char, columns in VARIANTS.get((name, is_bold), []):
            variant_hash = hash(tuple(columns)) & ((2 << 32) - 1)
            maybe_char_info = cls.update_automaton(
                char, columns, automaton, extra_info=f"{name}-{'bold' if is_bold else 'normal'}-{char}-{variant_hash:x}"
            )
            if maybe_char_info:
                accepting_states.append(maybe_char_info)
        initial_padding = max(char_info.left_padding for char_info in accepting_states)
        if not cls.is_mono:
            initial_padding = 2
        final_padding = max(char_info.right_padding for char_info in accepting_states)
        return cls(
            name,
            cls.get_space_width(image, width, is_bold, **kwargs),
            height,
            base,
            automaton,
            initial_padding,
            final_padding,
            is_bold,
            italic_offsets,
            group,
            max_cut_bottom,
            max_cut_top,
            display_name,
        )

    @classmethod
    @abstractmethod
    def get_input_columns(
        cls,
        image: SimpleImage,
        width: int,
        italic_offsets: set[int],
        is_bold: bool,
    ) -> Iterator[list[int]]:
        pass

    @classmethod
    @abstractmethod
    def get_space_width(cls, image: SimpleImage, width: int, is_bold: bool, **kwargs: Any) -> int:
        pass

    @classmethod
    def update_automaton(
        cls,
        char: str,
        columns: list[int],
        automaton: FSA,
        extra_info: str | None = None,
    ) -> CharInfo | None:
        x = 0
        while columns[x] == 0:
            x += 1
        left_padding = cls.base_left_padding + x
        x = len(columns) - 1
        while columns[x] == 0:
            x -= 1
        right_padding = cls.base_right_padding + len(columns) - 1 - x
        state = automaton
        from_column = left_padding - cls.base_left_padding
        if cls.is_mono:
            to_column = len(columns)
        else:
            to_column = len(columns) + (cls.base_right_padding if right_padding >= 0 else 0)
        for i, column in enumerate(columns[from_column:to_column]):
            state = state.setdefault(column, {})
            if right_padding > 1 and i >= len(columns) - right_padding - 1:
                if ACCEPT not in state or char in ("O", "l"):
                    state[ACCEPT] = CharInfo(char, left_padding, len(columns) - i, extra_info)
        # “ and ” look the same as " in some sizes,
        # but we prefer O to 0 and l to 1 when they look the same
        if ACCEPT not in state or char in ("O", "l"):
            state[ACCEPT] = CharInfo(char, left_padding, right_padding, extra_info)
            return state[ACCEPT]


class MonospaceFont(Font):
    @classmethod
    def get_input_columns(
        cls,
        image: SimpleImage,
        width: int,
        italic_offsets: set[int],
        is_bold: bool,
    ) -> Iterator[list[int]]:
        height = image.height
        for i in itertools.count():
            if width * (i+1) > image.width:
                return
            columns = [
                get_column(width * i + j, 0, image, height, italic_offsets)
                for j in range(width)
            ]
            if is_bold:
                columns = [column1 | column2 for column1, column2 in zip([0] + columns, columns + [0])]
            yield columns

    @classmethod
    def get_space_width(cls, image: SimpleImage, width: int, is_bold: bool, **kwargs: Any) -> int:
        return width + 1 if is_bold else width


@dataclass
class ProportionalFont(Font):
    is_mono: bool = False
    separator: int = 8
    prev_column_mask: int = 28
    base_left_padding: int = -1
    base_right_padding: int = -1
    skip_chars = ["1"]  # hard to distinguish from "l"

    @classmethod
    def get_input_columns(
        cls,
        image: SimpleImage,
        width: int,
        italic_offsets: set[int],
        is_bold: bool,
    ) -> Iterator[list[int]]:
        height = image.height
        columns = []
        first_column = True
        previous_column = 0
        for x in range(image.width):
            column = get_column(x, 0, image, height, set())
            if not first_column and column == cls.separator and previous_column & cls.prev_column_mask == 0:
                yield columns
                columns = []
                first_column = True
            else:
                columns.append(column)
                if column != 0:
                    first_column = False
            previous_column = column
        if columns:
            yield columns

    @classmethod
    def get_space_width(cls, image: SimpleImage, width: int, is_bold: bool, space_width: int = 0, **kwargs: Any) -> int:
        return space_width


def get_column(
    x0: int, y0: int, image: SimpleImage, height: int, italic_offsets: set[int], cut_bottom: int = 0, cut_top: int = 0
) -> int:
    bitmask = 0
    italic_offset = len(italic_offsets)
    for y in range(y0, y0 + height):
        if y - y0 in italic_offsets:
            italic_offset -= 1
        bitmask <<= 1
        if (x0 + italic_offset, y) in image.pixels:
            bitmask += 1
    return cut_column(bitmask, height, cut_bottom, cut_top)


def cut_column(column: int, height: int, cut_bottom: int = 0, cut_top: int = 0):
    bitmask = column
    if cut_bottom:
        bitmask = bitmask & -(1 << cut_bottom)
    if cut_top:
        bitmask = bitmask & ((1 << height - cut_top) - 1)
    return bitmask


def get_pixels_from_columns(columns: list[int], height: int, x0: int, y0: int, italic_offsets: set[int]) -> set[Pixel]:
    pixels = set()
    for x, column in enumerate(columns, start=x0):
        italic_offset = 0
        for y in range(y0 + height - 1, y0 - 1, -1):
            if column == 0:
                break
            if column & 1 == 1:
                pixels.add(Pixel(x + italic_offset, y))
            if y - y0 in italic_offsets:
                italic_offset += 1
            column >>= 1
    return pixels


ALL_FONTS: list[Font] = [
    MonospaceFont.from_file(
        file_path_context_manager=as_file(files(parse_qwantz).joinpath(f'img/regular{size}.png')),
        name=name,
        width=width,
        italic_offsets=set(),
        is_bold=is_bold,
        group=f'LC{size}',
        max_cut_bottom=max_cut_bottom,
        max_cut_top=max_cut_top,
        display_name=display_name,
    )
    for size, name, width, max_cut_bottom, max_cut_top, display_name in FONT_SIZES
    for is_bold in (False, True)
]

ALL_FONTS.append(
    MonospaceFont.from_file(
        file_path_context_manager=as_file(files(parse_qwantz).joinpath(f'img/italic13.png')),
        name='Italic',
        width=8,
        italic_offsets={3, 5, 9, 11},
        is_bold=False,
        group='LC13',
        max_cut_bottom=0,
        max_cut_top=0,
    )
)

ALL_FONTS.append(
    ProportionalFont.from_file(
        file_path_context_manager=as_file(files(parse_qwantz).joinpath(f'img/serif13.png')),
        name='Serif',
        width=0,
        italic_offsets=set(),
        is_bold=False,
        group='TNR13',
        max_cut_bottom=0,
        max_cut_top=0,
        space_width=3,
        display_name='serif'
    )
)
